// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

//! Fayalite Simulation

use crate::{
    bundle::{BundleField, BundleType},
    expr::{
        target::{
            Target, TargetBase, TargetPathArrayElement, TargetPathBundleField, TargetPathElement,
        },
        ExprEnum,
    },
    intern::{Intern, Interned, Memoize},
    module::{
        AnnotatedModuleIO, Block, ModuleBody, NormalModuleBody, Stmt, StmtConnect, StmtDeclaration,
        StmtFormal, StmtIf, StmtInstance, StmtMatch, StmtReg, StmtWire,
    },
    prelude::*,
    sim::interpreter::{
        Insn, InsnField, InsnFieldKind, InsnFieldType, Insns, InsnsBuilding, InsnsBuildingDone,
        SlotDebugData, SmallUInt, StatePartIndex, StatePartIndexMap, StatePartIndexRange,
        StatePartKind, StatePartKindBigSlots, StatePartKindSmallSlots, StatePartLayout,
        StatePartLen, StatePartsValue, TypeIndex, TypeIndexRange, TypeLayout, TypeLen, TypeParts,
        MIN_BITS_FOR_NEEDING_BIG,
    },
};
use hashbrown::HashMap;
use num_bigint::BigInt;
use std::{collections::BTreeSet, fmt, marker::PhantomData, mem};

mod interpreter;

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CondBody {
    IfTrue {
        cond: CompiledExpr<Bool>,
    },
    IfFalse {
        cond: CompiledExpr<Bool>,
    },
    MatchArm {
        enum_expr: CompiledExpr<Enum>,
        variant_index: usize,
    },
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct Cond {
    body: CondBody,
    source_location: SourceLocation,
}

#[derive(PartialEq, Eq, Hash, Clone, Copy)]
enum InstantiatedModule {
    Base(Interned<Module<Bundle>>),
    Child {
        parent: Interned<InstantiatedModule>,
        instance: Interned<Instance<Bundle>>,
    },
}

impl InstantiatedModule {
    fn leaf_module(self) -> Interned<Module<Bundle>> {
        match self {
            InstantiatedModule::Base(base) => base,
            InstantiatedModule::Child { instance, .. } => instance.instantiated(),
        }
    }
    fn write_path(self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            InstantiatedModule::Base(base) => fmt::Debug::fmt(&base.name_id(), f),
            InstantiatedModule::Child { parent, instance } => {
                parent.write_path(f)?;
                write!(f, ".{}", instance.name_id())
            }
        }
    }
}

impl fmt::Debug for InstantiatedModule {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "InstantiatedModule(")?;
        self.write_path(f)?;
        write!(f, ": {})", self.leaf_module().name_id())
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct TargetInInstantiatedModule {
    instantiated_module: InstantiatedModule,
    target: Target,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledBundleField {
    offset: TypeIndex,
    ty: CompiledTypeLayout<CanonicalType>,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CompiledTypeLayoutBody {
    Scalar,
    Array {
        /// debug names are ignored, use parent's layout instead
        element: Interned<CompiledTypeLayout<CanonicalType>>,
    },
    Bundle {
        /// debug names are ignored, use parent's layout instead
        fields: Interned<[CompiledBundleField]>,
    },
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledTypeLayout<T: Type> {
    ty: T,
    layout: TypeLayout<InsnsBuildingDone>,
    body: CompiledTypeLayoutBody,
}

impl<T: Type> CompiledTypeLayout<T> {
    fn with_prefixed_debug_names(self, prefix: &str) -> Self {
        let Self { ty, layout, body } = self;
        Self {
            ty,
            layout: layout.with_prefixed_debug_names(prefix),
            body,
        }
    }
    fn get(ty: T) -> Self {
        #[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
        struct MyMemoize;
        impl Memoize for MyMemoize {
            type Input = CanonicalType;
            type InputOwned = CanonicalType;
            type Output = CompiledTypeLayout<CanonicalType>;

            fn inner(self, input: &Self::Input) -> Self::Output {
                match input {
                    CanonicalType::UInt(_)
                    | CanonicalType::SInt(_)
                    | CanonicalType::Bool(_)
                    | CanonicalType::Enum(_)
                    | CanonicalType::AsyncReset(_)
                    | CanonicalType::SyncReset(_)
                    | CanonicalType::Reset(_)
                    | CanonicalType::Clock(_) => {
                        let mut layout = TypeLayout::empty();
                        let debug_data = SlotDebugData {
                            name: Interned::default(),
                            ty: *input,
                        };
                        layout.big_slots = StatePartLayout::scalar(debug_data);
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Scalar,
                        }
                    }
                    CanonicalType::Array(array) => {
                        let mut layout = TypeLayout::empty();
                        let element = CompiledTypeLayout::get(array.element()).intern_sized();
                        for index in 0..array.len() {
                            layout.allocate(
                                &element
                                    .layout
                                    .with_prefixed_debug_names(&format!("[{index}]")),
                            );
                        }
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Array { element },
                        }
                    }
                    CanonicalType::Bundle(bundle) => {
                        let mut layout = TypeLayout::empty();
                        let fields = bundle
                            .fields()
                            .iter()
                            .map(
                                |BundleField {
                                     name,
                                     flipped: _,
                                     ty,
                                 }| {
                                    let ty = CompiledTypeLayout::get(*ty);
                                    let offset = layout
                                        .allocate(
                                            &ty.layout
                                                .with_prefixed_debug_names(&format!(".{name}")),
                                        )
                                        .start();
                                    CompiledBundleField { offset, ty }
                                },
                            )
                            .collect();
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Bundle { fields },
                        }
                    }
                }
            }
        }
        let CompiledTypeLayout {
            ty: _,
            layout,
            body,
        } = MyMemoize.get_owned(ty.canonical());
        Self { ty, layout, body }
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledValue<T: Type> {
    layout: CompiledTypeLayout<T>,
    range: TypeIndexRange,
    write: Option<(CompiledTypeLayout<T>, TypeIndexRange)>,
}

impl<T: Type> CompiledValue<T> {
    fn map<U: Type>(
        self,
        mut f: impl FnMut(
            CompiledTypeLayout<T>,
            TypeIndexRange,
        ) -> (CompiledTypeLayout<U>, TypeIndexRange),
    ) -> CompiledValue<U> {
        let (layout, range) = f(self.layout, self.range);
        CompiledValue {
            layout,
            range,
            write: self.write.map(|(layout, range)| f(layout, range)),
        }
    }
}

impl CompiledValue<Enum> {
    fn add_discriminant_to_set(self, offset: TypeIndex, inputs: &mut SlotSet) {
        inputs.extend([self.range.offset(offset)]);
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Copy, Clone)]
struct CompiledExprDynIndex {
    index_slot: StatePartIndex<StatePartKindSmallSlots>,
    len: usize,
    stride: TypeLen,
}

impl CompiledExprDynIndex {
    fn offsets(self) -> impl Iterator<Item = TypeIndex> {
        (0..self.len.try_into().unwrap()).map(move |index| TypeIndex {
            small_slots: StatePartIndex {
                value: self
                    .stride
                    .small_slots
                    .value
                    .checked_mul(index)
                    .expect("array too big"),
                _phantom: PhantomData,
            },
            big_slots: StatePartIndex {
                value: self
                    .stride
                    .big_slots
                    .value
                    .checked_mul(index)
                    .expect("array too big"),
                _phantom: PhantomData,
            },
        })
    }
    fn for_each_offset(dyn_indexes: &[CompiledExprDynIndex], mut f: impl FnMut(TypeIndex)) {
        fn helper(
            dyn_indexes: &[CompiledExprDynIndex],
            offset: TypeIndex,
            f: &mut impl FnMut(TypeIndex),
        ) {
            if let [next, rest @ ..] = dyn_indexes {
                for next_offset in next.offsets() {
                    helper(
                        rest,
                        TypeIndex {
                            small_slots: StatePartIndex {
                                value: next_offset
                                    .small_slots
                                    .value
                                    .checked_add(offset.small_slots.value)
                                    .expect("array too big"),
                                _phantom: PhantomData,
                            },
                            big_slots: StatePartIndex {
                                value: next_offset
                                    .big_slots
                                    .value
                                    .checked_add(offset.big_slots.value)
                                    .expect("array too big"),
                                _phantom: PhantomData,
                            },
                        },
                        f,
                    );
                }
            } else {
                f(offset);
            }
        }
        helper(
            dyn_indexes,
            TypeIndex {
                small_slots: StatePartIndex {
                    value: 0,
                    _phantom: PhantomData,
                },
                big_slots: StatePartIndex {
                    value: 0,
                    _phantom: PhantomData,
                },
            },
            &mut f,
        );
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Copy, Clone)]
struct CompiledExpr<T: Type> {
    static_part: CompiledValue<T>,
    dyn_indexes: Interned<[CompiledExprDynIndex]>,
}

impl<T: Type> From<CompiledValue<T>> for CompiledExpr<T> {
    fn from(static_part: CompiledValue<T>) -> Self {
        Self {
            static_part,
            dyn_indexes: Interned::default(),
        }
    }
}

impl<T: Type> CompiledExpr<T> {
    fn map_ty<U: Type>(self, mut f: impl FnMut(T) -> U) -> CompiledExpr<U> {
        let Self {
            static_part,
            dyn_indexes,
        } = self;
        CompiledExpr {
            static_part: static_part.map(|CompiledTypeLayout { ty, layout, body }, range| {
                (
                    CompiledTypeLayout {
                        ty: f(ty),
                        layout,
                        body,
                    },
                    range,
                )
            }),
            dyn_indexes,
        }
    }
    fn add_target_without_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part,
            dyn_indexes,
        } = self;
        CompiledExprDynIndex::for_each_offset(&dyn_indexes, |offset| {
            inputs.extend([static_part.range.offset(offset)]);
        });
    }
    fn add_target_and_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part: _,
            dyn_indexes,
        } = self;
        self.add_target_without_indexes_to_set(inputs);
        inputs.extend(dyn_indexes);
    }
}

impl CompiledExpr<Enum> {
    fn add_discriminant_and_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part,
            dyn_indexes,
        } = self;
        CompiledExprDynIndex::for_each_offset(&dyn_indexes, |offset| {
            static_part.add_discriminant_to_set(offset, inputs);
        });
        inputs.extend(dyn_indexes);
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
struct SlotSet(TypeParts<Self>);

impl SlotSet {
    fn is_empty(&self) -> bool {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.is_empty() && big_slots.is_empty()
    }
}

impl StatePartsValue for SlotSet {
    type Value<K: StatePartKind> = BTreeSet<StatePartIndex<K>>;
}

impl Extend<StatePartIndex<StatePartKindSmallSlots>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindSmallSlots>>>(&mut self, iter: T) {
        self.0.small_slots.extend(iter);
    }
}

impl Extend<StatePartIndex<StatePartKindBigSlots>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindBigSlots>>>(&mut self, iter: T) {
        self.0.big_slots.extend(iter);
    }
}

impl<K: StatePartKind> Extend<StatePartIndexRange<K>> for SlotSet
where
    Self: Extend<StatePartIndex<K>>,
{
    fn extend<T: IntoIterator<Item = StatePartIndexRange<K>>>(&mut self, iter: T) {
        self.extend(iter.into_iter().flat_map(|v| v.iter()));
    }
}

impl Extend<TypeIndexRange> for SlotSet {
    fn extend<T: IntoIterator<Item = TypeIndexRange>>(&mut self, iter: T) {
        iter.into_iter().for_each(
            |TypeIndexRange {
                 small_slots,
                 big_slots,
             }| {
                self.extend(small_slots.iter());
                self.extend(big_slots.iter());
            },
        )
    }
}

impl Extend<CompiledExprDynIndex> for SlotSet {
    fn extend<T: IntoIterator<Item = CompiledExprDynIndex>>(&mut self, iter: T) {
        self.extend(iter.into_iter().map(|v| v.index_slot));
    }
}

impl Extend<CondBody> for SlotSet {
    fn extend<T: IntoIterator<Item = CondBody>>(&mut self, iter: T) {
        iter.into_iter().for_each(|cond_body| match cond_body {
            CondBody::IfTrue { cond } | CondBody::IfFalse { cond } => {
                cond.add_target_and_indexes_to_set(self);
            }
            CondBody::MatchArm {
                enum_expr,
                variant_index,
            } => enum_expr.add_discriminant_and_indexes_to_set(self),
        })
    }
}

impl Extend<Cond> for SlotSet {
    fn extend<T: IntoIterator<Item = Cond>>(&mut self, iter: T) {
        self.extend(iter.into_iter().map(|v| v.body))
    }
}

#[derive(Debug)]
struct Assignment {
    inputs: SlotSet,
    conditions: Interned<[Cond]>,
    insns: Vec<Insn>,
}

#[derive(Debug, Default)]
struct StatePartAssignments<K: StatePartKind> {
    written_slot_to_assignment_indexes_map: StatePartIndexMap<K, Vec<usize>>,
}

#[derive(Debug, Default)]
struct SlotAssignments {
    assignments: Vec<Assignment>,
    parts: TypeParts<Self>,
}

impl StatePartsValue for SlotAssignments {
    type Value<K: StatePartKind> = StatePartAssignments<K>;
}

#[derive(Debug)]
pub struct Compiler {
    insns: Insns<InsnsBuilding>,
    base_module: Interned<Module<Bundle>>,
    modules: HashMap<InstantiatedModule, CompiledModule>,
    compiled_values: HashMap<TargetInInstantiatedModule, CompiledValue<CanonicalType>>,
    compiled_exprs: HashMap<(Interned<[Cond]>, Expr<CanonicalType>), CompiledExpr<CanonicalType>>,
    compiled_exprs_to_values:
        HashMap<(Interned<[Cond]>, CompiledExpr<CanonicalType>), CompiledValue<CanonicalType>>,
    slots_assignments: SlotAssignments,
}

impl Compiler {
    pub fn new(base_module: Interned<Module<Bundle>>) -> Self {
        Self {
            insns: Insns::new(),
            base_module,
            modules: HashMap::new(),
            compiled_values: HashMap::new(),
            compiled_exprs: HashMap::new(),
            compiled_exprs_to_values: HashMap::new(),
            slots_assignments: SlotAssignments::default(),
        }
    }
    fn compile_value(
        &mut self,
        target: TargetInInstantiatedModule,
    ) -> CompiledValue<CanonicalType> {
        if let Some(&retval) = self.compiled_values.get(&target) {
            return retval;
        }
        let retval = match target.target {
            Target::Base(base) => {
                let unprefixed_layout = CompiledTypeLayout::get(base.canonical_ty());
                let layout = unprefixed_layout.with_prefixed_debug_names(&format!(
                    "{:?}.{:?}",
                    target.instantiated_module,
                    base.target_name()
                ));
                let range = self.insns.allocate_variable(&layout.layout);
                let write = match *base {
                    TargetBase::ModuleIO(_)
                    | TargetBase::MemPort(_)
                    | TargetBase::Wire(_)
                    | TargetBase::Instance(_) => None,
                    TargetBase::Reg(_) => {
                        let write_layout = unprefixed_layout.with_prefixed_debug_names(&format!(
                            "{:?}.{:?}$next",
                            target.instantiated_module,
                            base.target_name()
                        ));
                        Some((
                            write_layout,
                            self.insns.allocate_variable(&write_layout.layout),
                        ))
                    }
                };
                CompiledValue {
                    range,
                    layout,
                    write,
                }
            }
            Target::Child(target_child) => {
                let parent = self.compile_value(TargetInInstantiatedModule {
                    instantiated_module: target.instantiated_module,
                    target: *target_child.parent(),
                });
                match *target_child.path_element() {
                    TargetPathElement::BundleField(TargetPathBundleField { name }) => {
                        parent.map(|layout, range| {
                            let CompiledTypeLayout {
                                ty: CanonicalType::Bundle(bundle),
                                layout: _,
                                body: CompiledTypeLayoutBody::Bundle { fields },
                            } = layout
                            else {
                                unreachable!();
                            };
                            let field_index = bundle.name_indexes()[&name];
                            (
                                fields[field_index].ty,
                                range.slice(TypeIndexRange::new(
                                    fields[field_index].offset,
                                    fields[field_index].ty.layout.len(),
                                )),
                            )
                        })
                    }
                    TargetPathElement::ArrayElement(TargetPathArrayElement { index }) => parent
                        .map(|layout, range| {
                            let CompiledTypeLayoutBody::Array { element } = layout.body else {
                                unreachable!();
                            };
                            (*element, range.index_array(element.layout.len(), index))
                        }),
                    TargetPathElement::DynArrayElement(_) => unreachable!(),
                }
            }
        };
        self.compiled_values.insert(target, retval);
        retval
    }
    fn compiled_expr_to_value(
        &mut self,
        conditions: Interned<[Cond]>,
        expr: CompiledExpr<CanonicalType>,
    ) -> CompiledValue<CanonicalType> {
        if let Some(&retval) = self.compiled_exprs_to_values.get(&(conditions, expr)) {
            return retval;
        }
        let CompiledExpr {
            static_part: mut retval,
            dyn_indexes,
        } = expr;
        for CompiledExprDynIndex {
            index_slot,
            len,
            stride,
        } in dyn_indexes
        {
            todo!();
        }
        self.compiled_exprs_to_values
            .insert((conditions, expr), retval);
        retval
    }
    fn add_assignment(
        &mut self,
        conditions: Interned<[Cond]>,
        insns: impl IntoIterator<Item = Insn>,
    ) {
        let insns = Vec::from_iter(insns);
        let assignment_index = self.slots_assignments.assignments.len();
        let mut inputs = SlotSet::default();
        for insn in &insns {
            for InsnField { ty, kind } in insn.fields() {
                match (kind, ty) {
                    (InsnFieldKind::Input, InsnFieldType::SmallSlot(&small_slot)) => {
                        inputs.0.small_slots.insert(small_slot);
                    }
                    (InsnFieldKind::Input, InsnFieldType::BigSlot(&big_slot)) => {
                        inputs.0.big_slots.insert(big_slot);
                    }
                    (InsnFieldKind::Output, InsnFieldType::SmallSlot(&small_slot)) => self
                        .slots_assignments
                        .parts
                        .small_slots
                        .written_slot_to_assignment_indexes_map
                        .entry(small_slot)
                        .or_insert_with(Vec::new)
                        .push(assignment_index),
                    (InsnFieldKind::Output, InsnFieldType::BigSlot(&big_slot)) => self
                        .slots_assignments
                        .parts
                        .big_slots
                        .written_slot_to_assignment_indexes_map
                        .entry(big_slot)
                        .or_insert_with(Vec::new)
                        .push(assignment_index),
                    (
                        _,
                        InsnFieldType::SmallUInt(_)
                        | InsnFieldType::SmallSInt(_)
                        | InsnFieldType::InternedBigInt(_)
                        | InsnFieldType::U8(_)
                        | InsnFieldType::USize(_)
                        | InsnFieldType::Empty(_),
                    )
                    | (InsnFieldKind::Immediate | InsnFieldKind::BranchTarget, _) => {}
                }
            }
        }
        self.slots_assignments.assignments.push(Assignment {
            inputs,
            conditions,
            insns,
        });
    }
    fn simple_nary_big_expr<const N: usize>(
        &mut self,
        conditions: Interned<[Cond]>,
        dest_ty: CanonicalType,
        inputs: [Expr<CanonicalType>; N],
        make_insns: impl FnOnce(
            StatePartIndex<StatePartKindBigSlots>,
            [StatePartIndex<StatePartKindBigSlots>; N],
        ) -> Vec<Insn>,
    ) -> CompiledValue<CanonicalType> {
        let inputs = inputs.map(|input| {
            let input = self.compile_expr(conditions, input);
            let input = self.compiled_expr_to_value(conditions, input);
            let TypeIndexRange {
                small_slots,
                big_slots,
            } = input.range;
            assert_eq!(small_slots.len.value, 0);
            assert_eq!(big_slots.len.value, 1);
            big_slots.start
        });
        let layout = CompiledTypeLayout::get(dest_ty);
        let range = self.insns.allocate_variable(&layout.layout);
        let TypeIndexRange {
            small_slots,
            big_slots,
        } = range;
        assert_eq!(small_slots.len.value, 0);
        assert_eq!(big_slots.len.value, 1);
        let retval = CompiledValue {
            layout,
            range,
            write: None,
        };
        self.add_assignment(conditions, make_insns(big_slots.start, inputs));
        retval
    }
    fn compile_expr(
        &mut self,
        conditions: Interned<[Cond]>,
        expr: Expr<CanonicalType>,
    ) -> CompiledExpr<CanonicalType> {
        if let Some(&retval) = self.compiled_exprs.get(&(conditions, expr)) {
            return retval;
        }
        let mut cast_bit = |arg: Expr<CanonicalType>| {
            let src_signed = match Expr::ty(arg) {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => true,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(_) => unreachable!(),
                CanonicalType::Enum(_) => unreachable!(),
                CanonicalType::Bundle(_) => unreachable!(),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => false,
                CanonicalType::Clock(_) => false,
            };
            let dest_signed = match Expr::ty(expr) {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => true,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(_) => unreachable!(),
                CanonicalType::Enum(_) => unreachable!(),
                CanonicalType::Bundle(_) => unreachable!(),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => false,
                CanonicalType::Clock(_) => false,
            };
            self.simple_nary_big_expr(conditions, Expr::ty(expr), [arg], |dest, [src]| {
                match (src_signed, dest_signed) {
                    (false, false) | (true, true) => {
                        vec![Insn::Copy { dest, src }]
                    }
                    (false, true) => vec![Insn::CastToSInt {
                        dest,
                        src,
                        dest_width: 1,
                    }],
                    (true, false) => vec![Insn::CastToUInt {
                        dest,
                        src,
                        dest_width: 1,
                    }],
                }
            })
            .into()
        };
        let retval: CompiledExpr<_> = match *Expr::expr_enum(expr) {
            ExprEnum::UIntLiteral(expr) => self
                .simple_nary_big_expr(conditions, expr.ty().canonical(), [], |dest, []| {
                    vec![Insn::Const {
                        dest,
                        value: expr.to_bigint().intern_sized(),
                    }]
                })
                .into(),
            ExprEnum::SIntLiteral(expr) => self
                .simple_nary_big_expr(conditions, expr.ty().canonical(), [], |dest, []| {
                    vec![Insn::Const {
                        dest,
                        value: expr.to_bigint().intern_sized(),
                    }]
                })
                .into(),
            ExprEnum::BoolLiteral(expr) => self
                .simple_nary_big_expr(conditions, Bool.canonical(), [], |dest, []| {
                    vec![Insn::Const {
                        dest,
                        value: BigInt::from(expr).intern_sized(),
                    }]
                })
                .into(),
            ExprEnum::BundleLiteral(expr) => todo!(),
            ExprEnum::ArrayLiteral(expr) => todo!(),
            ExprEnum::EnumLiteral(expr) => todo!(),
            ExprEnum::Uninit(expr) => todo!(),
            ExprEnum::NotU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::NotU {
                            dest,
                            src,
                            width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::NotS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::NotS { dest, src }],
                )
                .into(),
            ExprEnum::NotB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::NotU {
                            dest,
                            src,
                            width: 1,
                        }]
                    },
                )
                .into(),
            ExprEnum::Neg(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::Neg { dest, src }],
                )
                .into(),
            ExprEnum::BitAndU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitAndS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitAndB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::AddU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Add { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::AddS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Add { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::SubU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| {
                        vec![Insn::SubU {
                            dest,
                            lhs,
                            rhs,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SubS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::SubS { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::MulU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Mul { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::MulS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Mul { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DivU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Div { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DivS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Div { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::RemU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Rem { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::RemS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Rem { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShlU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShl { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShlS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShl { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShrU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShr { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShrS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShr { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::FixedShlU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shl {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShlS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shl {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShrU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shr {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShrS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shr {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CmpLtB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeB(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLtU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeU(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLtS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeS(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CastUIntToUInt(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToUInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastUIntToSInt(expr) => todo!(),
            ExprEnum::CastSIntToUInt(expr) => self
                .simple_nary_big_expr(
                    conditions,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToUInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastSIntToSInt(expr) => todo!(),
            ExprEnum::CastBoolToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::FieldAccess(expr) => todo!(),
            ExprEnum::VariantAccess(expr) => todo!(),
            ExprEnum::ArrayIndex(expr) => todo!(),
            ExprEnum::DynArrayIndex(expr) => todo!(),
            ExprEnum::ReduceBitAndU(expr) => todo!(),
            ExprEnum::ReduceBitAndS(expr) => todo!(),
            ExprEnum::ReduceBitOrU(expr) => todo!(),
            ExprEnum::ReduceBitOrS(expr) => todo!(),
            ExprEnum::ReduceBitXorU(expr) => todo!(),
            ExprEnum::ReduceBitXorS(expr) => todo!(),
            ExprEnum::SliceUInt(expr) => todo!(),
            ExprEnum::SliceSInt(expr) => todo!(),
            ExprEnum::CastToBits(expr) => todo!(),
            ExprEnum::CastBitsTo(expr) => todo!(),
            ExprEnum::ModuleIO(expr) => todo!(),
            ExprEnum::Instance(expr) => todo!(),
            ExprEnum::Wire(expr) => todo!(),
            ExprEnum::Reg(expr) => todo!(),
            ExprEnum::MemPort(expr) => todo!(),
        };
        self.compiled_exprs.insert((conditions, expr), retval);
        retval
    }
    fn compile_connect(
        &mut self,
        parent_module: Interned<InstantiatedModule>,
        conditions: Interned<[Cond]>,
        lhs: Expr<CanonicalType>,
        mut rhs: Expr<CanonicalType>,
        source_location: SourceLocation,
    ) {
        if Expr::ty(lhs) != Expr::ty(rhs) || !Expr::ty(lhs).is_passive() {
            match Expr::ty(lhs) {
                CanonicalType::UInt(lhs_ty) => {
                    rhs = Expr::canonical(Expr::<UInt>::from_canonical(rhs).cast_to(lhs_ty));
                }
                CanonicalType::SInt(lhs_ty) => {
                    rhs = Expr::canonical(Expr::<SInt>::from_canonical(rhs).cast_to(lhs_ty));
                }
                CanonicalType::Bool(_) => unreachable!(),
                CanonicalType::Array(lhs_ty) => {
                    let CanonicalType::Array(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    assert_eq!(lhs_ty.len(), rhs_ty.len());
                    let lhs = Expr::<Array>::from_canonical(lhs);
                    let rhs = Expr::<Array>::from_canonical(rhs);
                    for index in 0..lhs_ty.len() {
                        self.compile_connect(
                            parent_module,
                            conditions,
                            lhs[index],
                            rhs[index],
                            source_location,
                        );
                    }
                    return;
                }
                CanonicalType::Enum(lhs_ty) => {
                    let CanonicalType::Enum(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    todo!("handle connect with different enum types");
                }
                CanonicalType::Bundle(lhs_ty) => {
                    let CanonicalType::Bundle(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    assert_eq!(lhs_ty.fields().len(), rhs_ty.fields().len());
                    let lhs = Expr::<Bundle>::from_canonical(lhs);
                    let rhs = Expr::<Bundle>::from_canonical(rhs);
                    for (
                        field_index,
                        (
                            BundleField {
                                name,
                                flipped,
                                ty: _,
                            },
                            rhs_field,
                        ),
                    ) in lhs_ty.fields().into_iter().zip(rhs_ty.fields()).enumerate()
                    {
                        assert_eq!(name, rhs_field.name);
                        assert_eq!(flipped, rhs_field.flipped);
                        let mut lhs_expr =
                            crate::expr::ops::FieldAccess::new_by_index(lhs, field_index).to_expr();
                        let mut rhs_expr =
                            crate::expr::ops::FieldAccess::new_by_index(rhs, field_index).to_expr();
                        if flipped {
                            mem::swap(&mut lhs_expr, &mut rhs_expr);
                        }
                        self.compile_connect(
                            parent_module,
                            conditions,
                            lhs_expr,
                            rhs_expr,
                            source_location,
                        );
                    }
                    return;
                }
                CanonicalType::AsyncReset(_) => unreachable!(),
                CanonicalType::SyncReset(_) => unreachable!(),
                CanonicalType::Reset(_) => unreachable!(),
                CanonicalType::Clock(_) => unreachable!(),
            }
        }
        let lhs = self.compile_expr(conditions, lhs);
        let rhs = self.compile_expr(conditions, rhs);
        let rhs = self.compiled_expr_to_value(conditions, rhs);
        todo!();
    }
    fn compile_block(
        &mut self,
        parent_module: Interned<InstantiatedModule>,
        block: Block,
        conditions: Interned<[Cond]>,
    ) {
        let Block { memories, stmts } = block;
        for memory in memories {
            todo!("implement memory");
        }
        for stmt in stmts {
            match stmt {
                Stmt::Connect(StmtConnect {
                    lhs,
                    rhs,
                    source_location,
                }) => self.compile_connect(parent_module, conditions, lhs, rhs, source_location),
                Stmt::Formal(StmtFormal {
                    kind,
                    clk,
                    pred,
                    en,
                    text,
                    source_location,
                }) => todo!("implement simulating formal statements"),
                Stmt::If(StmtIf {
                    cond,
                    source_location,
                    blocks: [then_block, else_block],
                }) => {
                    let cond = self
                        .compile_expr(conditions, Expr::canonical(cond))
                        .map_ty(Bool::from_canonical);
                    self.compile_block(
                        parent_module,
                        then_block,
                        Interned::from_iter(conditions.iter().copied().chain([Cond {
                            body: CondBody::IfTrue { cond },
                            source_location,
                        }])),
                    );
                    self.compile_block(
                        parent_module,
                        else_block,
                        Interned::from_iter(conditions.iter().copied().chain([Cond {
                            body: CondBody::IfFalse { cond },
                            source_location,
                        }])),
                    );
                }
                Stmt::Match(StmtMatch {
                    expr,
                    source_location,
                    blocks,
                }) => {
                    let enum_expr = self
                        .compile_expr(conditions, Expr::canonical(expr))
                        .map_ty(Enum::from_canonical);
                    for (variant_index, block) in blocks.into_iter().enumerate() {
                        self.compile_block(
                            parent_module,
                            block,
                            Interned::from_iter(conditions.iter().copied().chain([Cond {
                                body: CondBody::MatchArm {
                                    enum_expr,
                                    variant_index,
                                },
                                source_location,
                            }])),
                        );
                    }
                }
                Stmt::Declaration(declaration) => match declaration {
                    StmtDeclaration::Wire(StmtWire { annotations, wire }) => {
                        self.compile_value(TargetInInstantiatedModule {
                            instantiated_module: *parent_module,
                            target: wire.into(),
                        });
                    }
                    StmtDeclaration::Reg(StmtReg { annotations, reg }) => {
                        self.compile_value(TargetInInstantiatedModule {
                            instantiated_module: *parent_module,
                            target: reg.into(),
                        });
                        todo!();
                    }
                    StmtDeclaration::Instance(StmtInstance {
                        annotations,
                        instance,
                    }) => {
                        let CompiledValue {
                            layout:
                                CompiledTypeLayout {
                                    ty: value_ty,
                                    layout: ty_layout,
                                    body: CompiledTypeLayoutBody::Bundle { fields },
                                },
                            range: value_range,
                            write: None,
                        } = self.compile_value(TargetInInstantiatedModule {
                            instantiated_module: *parent_module,
                            target: instance.into(),
                        })
                        else {
                            unreachable!();
                        };
                        let CompiledModule { module_io } = *self.compile_module(
                            InstantiatedModule::Child {
                                parent: parent_module,
                                instance: instance.intern_sized(),
                            }
                            .intern_sized(),
                        );
                        for (module_io, CompiledBundleField { offset, ty }) in
                            module_io.into_iter().zip(fields)
                        {
                            todo!();
                        }
                        todo!()
                    }
                },
            }
        }
    }
    fn compile_module(&mut self, module: Interned<InstantiatedModule>) -> &CompiledModule {
        let module_io = module
            .leaf_module()
            .module_io()
            .iter()
            .map(
                |&AnnotatedModuleIO {
                     annotations: _,
                     module_io,
                 }| {
                    self.compile_value(TargetInInstantiatedModule {
                        instantiated_module: *module,
                        target: Target::from(module_io),
                    })
                },
            )
            .collect();
        match module.leaf_module().body() {
            ModuleBody::Normal(NormalModuleBody { body }) => {
                self.compile_block(module, body, Interned::default());
            }
            ModuleBody::Extern(_extern_module_body) => {
                todo!("simulating extern module: {:?}", module);
            }
        }
        let hashbrown::hash_map::Entry::Vacant(entry) = self.modules.entry(*module) else {
            unreachable!("compiled same instantiated module twice");
        };
        entry.insert(CompiledModule { module_io })
    }
    pub fn compile(mut self) -> Compiled {
        self.compile_module(InstantiatedModule::Base(self.base_module).intern_sized());

        Compiled {
            insns: Insns::from(self.insns).intern_sized(),
            modules: self.modules,
        }
    }
}

#[derive(Debug)]
struct CompiledModule {
    module_io: Interned<[CompiledValue<CanonicalType>]>,
}

#[derive(Debug)]
pub struct Compiled {
    insns: Interned<Insns<InsnsBuildingDone>>,
    modules: HashMap<InstantiatedModule, CompiledModule>,
}

impl Compiled {
    pub fn new<T: BundleType>(module: Module<T>) -> Self {
        Compiler::new(module.canonical().intern()).compile()
    }
}

#[derive(Debug)]
pub struct Simulation {
    state: interpreter::State,
}
